/*$$FILE$$*/
/*$$VERSION$$*/
/*$$DATE$$*/
/*$$LICENSE$$*/


#ifndef CIFDEFS_H
#define CIFDEFS_H

#define    CIF_DDL_TEMP_ITEM_CATEGORY_ID                 "temp.category_id"
#define    CIF_DDL_TEMP_ITEM_ITEM                        "temp.name"

#define    CIF_DDL_NAME_CATEGORY_ID                      "_category.id"
#define    CIF_DDL_NAME_ITEM_NAME                        "_item.name"

#define    CIF_DDL_ITEM_DATA_TYPE                        "data_type"
#define    CIF_DDL_ITEM_CASE_SENSITIVE                   "case_sensitive"
#define    CIF_DDL_ITEM_STRUCTURE_INDEX                  "structure_index"
#define    CIF_DDL_ITEM_STRUCTURE_ORGANIZATION           "structure_organization"
#define    CIF_DDL_ITEM_STRUCTURE_DIMENSION              "structure_dimension"
#define    CIF_DDL_ITEM_STRUCTURE_NVALUES                "structure_nvalues"
#define    CIF_DDL_ITEM_CONSTRUCT_INDEX                  "construct_index"

#define    CIF_DDL_NON_CASE_SENSITIVE_TERM               "uchar"
#define    CIF_DDL_CODE_TERM                             "code"
#define    CIF_DDL_ID_NAME_TERM                          "idname"
#define    CIF_DDL_NAME_TERM                             "name"
#define    CIF_INTEGER_TERM                              "int"
#define    CIF_INT_TERM                                  "integer"
#define    CIF_DOUBLE_TERM                               "double"
#define    CIF_FLOAT_TERM                                "float"
#define    CIF_REAL_TERM                                 "real"
#define    CIF_IMPLICIT_TERM                             "implicit"
#define    CIF_YES_TERM                                  "yes"
#define    CIF_NO_TERM                                   "no"
#define    CIF_RELATED_ITEM_FUNCTION_ASSOCIATED_ESD      "associated_esd"
#define    CIF_RELATED_ITEM_FUNCTION_ASSOCIATED_VALUE    "associated_value"

#define    CIF_DDL_CATEGORY_LINKAGE                      "item_linked"
#define    CIF_DDL_ITEM_LINKAGE_PARENT                   "parent_name"
#define    CIF_DDL_ITEM_LINKAGE_CHILD                    "child_name"

#define    CIF_RESERVED_KEYWORD_GLOBAL                   "global_"
#define    CIF_RESERVED_KEYWORD_DATA                     "data_"
#define    CIF_RESERVED_KEYWORD_SAVE                     "save_"
#define    CIF_RESERVED_KEYWORD_LOOP                     "loop_"
#define    CIF_RESERVED_KEYWORD_INCLUDE                  "include_"

#endif

